/* FILE: uniformvectorfieldinit.cc     -*-Mode: c++-*-
 *
 * Uniform vector field initializer, derived from Oxs_VectorFieldInit
 * class.
 *
 */

#include "oc.h"
#include "nb.h"

#include "uniformvectorfieldinit.h"

// Oxs_Ext registration support
OXS_EXT_REGISTER(Oxs_UniformVectorFieldInit);

/* End includes */


// Constructor
Oxs_UniformVectorFieldInit::Oxs_UniformVectorFieldInit(
  const char* name,     // Child instance id
  Oxs_Director* newdtr, // App director
  Tcl_Interp* safe_interp, // Safe interpreter
  const char* argstr)   // MIF input block parameters
  : Oxs_VectorFieldInit(name,newdtr,safe_interp,argstr)
{
  // Process arguments
  CheckInitValueParamCount("vector",3);
  const vector<string>& vstr = *FindInitValue("vector");
  vec.x=Nb_Atof(vstr[0].c_str());
  vec.y=Nb_Atof(vstr[1].c_str());
  vec.z=Nb_Atof(vstr[2].c_str());
  DeleteInitValue("vector");

  const vector<string>* nstr = FindInitValue("norm");
  if(nstr!=NULL) {
    CheckInitValueParamCount("norm",1);
    vec.SetMag(Nb_Atof((*nstr)[0].c_str()));
    DeleteInitValue("norm");
  }
  VerifyAllInitArgsUsed();
}

BOOL Oxs_UniformVectorFieldInit::Init()
{ return 1; }


void
Oxs_UniformVectorFieldInit::FillMeshValue
(const Oxs_Mesh* mesh,
 Oxs_MeshValue<ThreeVector>& array) const
{
  array.AdjustSize(mesh);
  UINT4m i,size=mesh->Size();
  for(i=0;i<size;i++) array[i]=vec;
}
