# FILE: oxsii.tcl
#
# The OOMMF eXtensible Solver Interative Interface

# Support libraries
package require Oc 1.1
package require Oxs 1.2
package require Net

Oc_IgnoreTermLoss  ;# Try to keep going, even if controlling terminal
## goes down.

# Application description boilerplate
Oc_Main SetAppName Oxsii
Oc_Main SetVersion 1.2.0.1
regexp \\\044Date:(.*)\\\044 {$Date: 2001/01/20 03:34:30 $} _ date
Oc_Main SetDate [string trim $date]
Oc_Main SetAuthor [Oc_Person Lookup dgp]
Oc_Main SetHelpURL [Oc_Url FromFilename [file join [file dirname \
        [file dirname [file dirname [Oc_DirectPathname [info \
        script]]]]] doc userguide userguide\
        OOMMF_eXtensible_Solver_Int.html]]

# Process command line options
Oc_CommandLine Option restart {
	{flag {expr {![catch {expr {$flag && $flag}}]}} {= 0 or 1}}
    } {
	global restart_option; set restart_option $flag
} {1 => use <basename>.log file to restart simulation}
set restart_option 0

Oc_CommandLine Parse $argv

##########################################################################
# Define the GUI of this app to be displayed remotely by clients of the
# GeneralServices protocol.  Return $gui in response to the GetGui message.
##########################################################################
set gui {
    package require Oc 1.1
    package require Tk
    package require Ow
    wm withdraw .
}
append gui "[list Oc_Main SetAppName [Oc_Main GetAppName]]\n"
append gui "[list Oc_Main SetVersion [Oc_Main GetVersion]]\n"
append gui "[list Oc_Main SetPid [pid]]\n"
append gui {

regexp \\\044Date:(.*)\\\044 {$Date: 2001/01/20 03:34:30 $} _ date
Oc_Main SetDate [string trim $date]

# This won't cross different OOMMF installations nicely
Oc_Main SetAuthor [Oc_Person Lookup donahue]

}
# This might not cross nicely either
append gui {Oc_Main SetHelpURL [Oc_Url FromFilename }
append gui [list [file join [file dirname [file dirname [file dirname \
	[Oc_DirectPathname [info script]]]]] doc userguide userguide \
        OOMMF_eXtensible_Solver_Int.html]]]\n
append gui {
# Enable display of remote errors
share errorstate
trace variable errorstate w {DisplayError ;# }
proc DisplayError {} {
    global errorstate errorInfo errorCode
    if {[string match {} $errorstate]} {return}
    foreach {s t m errorInfo errorCode} $errorstate {break}
    Oc_Log Log "$m\n($s $t)" error
}

set menubar .mb
foreach {fmenu hmenu} [Ow_MakeMenubar . $menubar File Help] {}
$fmenu add command -label "Load..." \
	-command [list LoadButton $fmenu "Load..."] -underline 0
$fmenu add command -label "Show Console" -command { console show } -underline 0
$fmenu add command -label "Close Interface" -command { exit } -underline 0
$fmenu add command -label "Exit Oxsii" -command { appexit; exit } -underline 1
Ow_StdHelpMenu $hmenu
proc LoadButton { btn item } {
    if { [string match disabled [$btn entrycget $item -state]] } {
        return
    }
    Ow_FileDlg New dialog -callback LoadCallback \
	    -dialog_title "[Oc_Main GetInstanceName]: Load Problem" \
            -selection_title "Load MIF File..." \
            -select_action_id LOAD \
	    -filter *.mif \
	    -file_must_exist 1 \
	    -delete_callback_arg [list $btn entryconfigure $item -state normal]
    $btn entryconfigure $item -state disabled
}
proc LoadCallback { widget actionid args } {
    global problem
    if {[string match DELETE $actionid]} {
	# This really is the way it has to be!  The quoting on $args
	# is just plain weird.
        eval [join $args]
        return
    }
    if {[string match LOAD $actionid]} {
	set problem [$widget GetFilename]
    } else {
        return "ERROR (proc LoadCallBack): Invalid actionid: $actionid"
    }
    # The following [return] must be here so the dialog will close 
    return
}

set menuwidth [Ow_GuessMenuWidth $menubar]
set brace [canvas .brace -width $menuwidth -height 0 -borderwidth 0 \
        -highlightthickness 0]
pack $brace -side top

wm title . [Oc_Main GetInstanceName]
wm iconname . [wm title .]
Ow_SetIcon .
wm deiconify .

if {[package vcompare [package provide Tk] 8] >= 0 \
        && [string match windows $tcl_platform(platform)]} {
    # Windows doesn't size Tcl 8.0 menubar cleanly
    Oc_DisableAutoSize .
    wm geometry . "${menuwidth}x0"
    update
    wm geometry . {}
    Oc_EnableAutoSize .
    update
}

set interface_state disabled
share interface_state

set bf [frame .buttons]
pack [button $bf.reload -text Reload -command {set problem $problem} \
	-state $interface_state] -fill x -side left
pack [button $bf.reset -text Reset -command {set status Initializing...} \
	-state $interface_state] -fill x -side left
pack [button $bf.run -text Run -command {set status Run} \
	-state $interface_state] -fill x -side left
pack [button $bf.relax -text Relax -command {set status Relax} \
	-state $interface_state] -fill x -side left
pack [button $bf.step -text Step -command {set status Step} \
	-state $interface_state] -fill x -side left
pack [button $bf.pause -text Pause -command {set status Pause} \
	-state $interface_state] -fill x -side left

pack $bf -side top -fill both -expand 1

trace variable interface_state w {
    foreach _ [winfo children .buttons] {
	$_ configure -state [uplevel #0 set interface_state];
    } ;# }

share problem
set probframe [frame .problem ]
pack [label $probframe.l -text Problem: -anchor w] -side left -fill x
pack [label $probframe.val -textvariable problem -anchor e] -side right -fill x
pack $probframe -side top -fill x 

set mf [frame .monitor]

share status
set statframe [frame $mf.status]
pack [label $statframe.l -text Status: -anchor w] -side left -fill x
pack [label $statframe.val -textvariable status -anchor e] -side right -fill x
pack $statframe -side top -fill x

share stage
set stageframe [frame $mf.stage]
pack [label $stageframe.l -text Stage: -anchor w -width 18] -side left -fill x
pack [label $stageframe.val -textvariable stage -anchor e] -side right -fill x
pack $stageframe -side top -fill x

share step
set stepframe [frame $mf.step]
pack [label $stepframe.l -text Step: -anchor w] -side left -fill x
pack [label $stepframe.val -textvariable step -anchor e] -side right -fill x
pack $stepframe -side top -fill x

pack $mf -side top -fill both -expand 1

set oframe [frame .sopanel]

set opframe [frame $oframe.outputs]
pack [label $opframe.l -text Output -relief groove] -side top -fill x
Ow_ListBox New oplb $opframe.lb -height 3 -variable oplbsel
pack [$oplb Cget -winpath] -side top -fill x
share opAssocList
trace variable opAssocList w { uplevel #0 {
    catch {unset opArray}
    array set opArray $opAssocList
    set opList [lsort [array names opArray]] } ;# }
trace variable opList w {
    $oplb Remove 0 end
    eval [list $oplb Insert 0] $opList ;# }
trace variable oplbsel w {uplevel #0 {
    trace vdelete destList w {SetDestinationList; UpdateSchedule ;#}
    upvar 0 $opArray([lindex $oplbsel 0])List destList
    SetDestinationList
    trace variable destList w {SetDestinationList; UpdateSchedule ;#}
} ;# }
pack $opframe -side left -fill both

# The destination lists (Find a way to construct dynamically, yet
# efficiently):
share vectorFieldList
share DataTableList
set vectorFieldList [list]
set DataTableList [list]

set dframe [frame $oframe.destinations]
pack [label $dframe.l -text Destination -relief groove] -side top -fill x
Ow_ListBox New dlb $dframe.lb -height 3 -variable dlbsel
pack [$dlb Cget -winpath] -side top -fill x
proc SetDestinationList {} {
    global dlb destList
    $dlb Remove 0 end
    eval [list $dlb Insert 0] $destList
}
trace variable destList w {SetDestinationList; UpdateSchedule ;#}
pack $dframe -side left -fill both
	
set sframe [frame $oframe.schedule]
grid [label $sframe.l -text Schedule -relief groove] - -sticky new
grid columnconfigure $sframe 0 -pad 15
grid columnconfigure $sframe 1 -weight 1
grid [button $sframe.send -command \
	{remote Oxs_Output Send [lindex $oplbsel 0] [lindex $dlbsel 0]} \
	-text Send] - -sticky new
#
# FOR NOW JUST HACK IN THE EVENTS WE SUPPORT
#	eventually this may be driver-dependent
set events [list Step Stage]
set schedwidgets [list $sframe.send]
foreach event $events {
    set active [checkbutton $sframe.a$event -text $event -anchor w \
	    -variable Schedule---activeA($event)]
    $active configure -command [subst -nocommands \
		{Schedule Active [$active cget -variable] $event}]
    Ow_EntryBox New frequency $sframe.f$event -label every \
	    -autoundo 0 -valuewidth 4 \
	    -variable Schedule---frequencyA($event) \
	    -callback [list EntryCallback $event] \
	    -foreground Black -disabledforeground #a3a3a3 \
	    -valuetype posint -coloredits 1 -writethrough 0 \
	    -outer_frame_options "-bd 0"
    grid $active [$frequency Cget -winpath] -sticky nw
    lappend schedwidgets $active $frequency
}

grid rowconfigure $sframe [expr {[lindex [grid size $sframe] 1] - 1}] -weight 1
pack $sframe -side left -fill both -expand 1

proc EntryCallback {e w args} {
    upvar #0 [$w Cget -variable] var
    set var [$w Cget -value]
    Schedule Frequency [$w Cget -variable] $e
}
proc Schedule {x v e} {
    global oplbsel dlbsel
    upvar #0 $v value
    remote Oxs_Schedule Set [lindex $oplbsel 0] [lindex $dlbsel 0] $x $e $value
}

trace variable oplbsel w "UpdateSchedule ;#"
trace variable dlbsel w "UpdateSchedule ;#"
Oc_EventHandler Bindtags UpdateSchedule UpdateSchedule
proc UpdateSchedule {} {
    global oplbsel dlbsel schedwidgets opArray
    Oc_EventHandler Generate UpdateSchedule Reset
    set os [lindex $oplbsel 0]
    set ds [lindex $dlbsel 0]
    set state disabled
    if {[info exists opArray($os)]} {
	upvar #0 $opArray($os)List destList
	if {[lsearch -exact $destList $ds] >= 0} {
	    set state normal
	}
    }
    foreach _ $schedwidgets {
	if {[string match *EntryBox* $_]} {
#Oc_Log Log "$_ Configure -state $state" warning
	    $_ Configure -state $state
	} else {
	    $_ configure -state $state
	}
    }
    if {[string compare normal $state]} {return}
    regsub -all {:+} $os : os
    regsub -all "\[ \r\t\n]+" $os _ os
    regsub -all {:+} $ds : ds
    regsub -all "\[ \r\t\n]+" $ds _ ds
    global Schedule-$os-$ds-active
    global Schedule-$os-$ds-frequency
    global Schedule---frequencyA

    if {![info exists Schedule-$os-$ds-active]} {
	share Schedule-$os-$ds-active
	trace variable Schedule-$os-$ds-active w [subst { uplevel #0 {
	    catch {[list unset Schedule-$os-$ds-activeA]}
	    [list array set Schedule-$os-$ds-activeA] \[[list set \
	    Schedule-$os-$ds-active]] } ;# }]
    }

    if {![info exists Schedule-$os-$ds-frequency]} {
	share Schedule-$os-$ds-frequency
    } else {
	array set Schedule---frequencyA [set Schedule-$os-$ds-frequency]
    }

    # Reconfigure Schedule widgets
    foreach _ $schedwidgets {
	if {[regexp {.+[.]a([^.]+)$} $_ -> e]} {
	    $_ configure -variable Schedule-$os-$ds-activeA($e)
	}
    }

    # When entry boxes commit - write change to shared variable.
    trace variable Schedule---frequencyA w [subst { uplevel #0 {
	    [list set Schedule-$os-$ds-frequency] \
	    \[[list array get Schedule---frequencyA]]} ;# }]

    Oc_EventHandler New _ UpdateSchedule Reset \
	    [subst {trace vdelete Schedule---frequencyA w { uplevel #0 {
	    [list set Schedule-$os-$ds-frequency] \
	    \[[list array get Schedule---frequencyA]]} ;# }}] \
	    -oneshot 1

    # When shared variable changes - write change to entry box
    trace variable Schedule-$os-$ds-frequency w [subst { uplevel #0 {
	    [list array set Schedule---frequencyA] \[[list set \
	    Schedule-$os-$ds-frequency]] } ;# }]

    Oc_EventHandler New _ UpdateSchedule Reset [subst \
	    {[list trace vdelete Schedule-$os-$ds-frequency w] { uplevel #0 {
	    [list array set Schedule---frequencyA] \[[list set \
	    Schedule-$os-$ds-frequency]] } ;# }}] \
	    -oneshot 1
}
UpdateSchedule

pack $oframe -side top -fill x -expand 1

}
##########################################################################
# End of GUI script.
##########################################################################

##########################################################################
# Handle Tk
##########################################################################
if {[Oc_Main HasTk]} {
    wm withdraw .        ;# "." is just an empty window.
    package require Ow
    Ow_SetIcon .

    # Evaluate $gui?  (in a slave interp?) so that when this app
    # is run with Tk it presents locally the same interface it
    # otherwise exports to be displayed remotely?

} else {
    # Set up to write log messages to oommf/oxsii.errors.
    FileLogger SetFile \
	    [file join [Oc_Main GetOOMMFRootDir] oxsii.errors]
    Oc_Log SetLogHandler [list FileLogger Log] panic Oc_Log
    Oc_Log SetLogHandler [list FileLogger Log] error Oc_Log
    Oc_Log SetLogHandler [list FileLogger Log] warning Oc_Log
    Oc_Log SetLogHandler [list FileLogger Log] info Oc_Log
    Oc_Log SetLogHandler [list FileLogger Log] panic
    Oc_Log SetLogHandler [list FileLogger Log] error
    Oc_Log SetLogHandler [list FileLogger Log] warning
    Oc_Log SetLogHandler [list FileLogger Log] info

    # Send error information to all gui clients 
    Oc_EventHandler New _ FileLogger Log {
	set errorstate [list %src %type %msg %errorinfo %errorcode]
	set errorstate ""
    }
}

##########################################################################
# Class representing an output destination thread.  Maintains global
# lists for each protocol.
##########################################################################
Oc_Class Oxs_Destination {
    array common directory
    array common index
    const public variable name
    const public variable protocol

    Constructor {thread} {
	set pid [$thread Cget -pid]
        set name [$thread Cget -alias]<$pid>
	if {[info exists index($name)]} {
	    return -code error "Duplicate destination name"
	}
	set index($name) $thread
	set protocol [$thread Protocol]
	upvar #0 ${protocol}List list
	lappend list $name
	set list [lsort $list]
	set directory($name) $this
	foreach o [Oxs_Output Instances] {
	    if {[string compare $protocol [$o Protocol]] == 0} {
		Oxs_Schedule New _ -output [$o Cget -name] \
			-destination $name
	    }
	}
	if {[string compare $protocol DataTable] == 0} {
	    Oxs_Schedule New _ -output DataTable -destination $name
	}
	Oc_EventHandler New _ $thread Delete [list $this Delete] \
		-groups [list $this]
    }
    Destructor {
	Oc_EventHandler Generate $this Delete
	Oc_EventHandler DeleteGroup $this
	upvar #0 ${protocol}List list
	set idx [lsearch -exact $list $name]
	set list [lreplace $list $idx $idx]
	catch {unset index($name)}
    }
    proc Lookup {n} {
	return $directory($n)
    }
    proc Thread {n} {
	return $index($n)
    }
}
# Create a new Oxs_Destination for each Net_Thread that becomes Ready
Oc_EventHandler New _ Net_Thread Ready [list Oxs_Destination New _ %object]

##########################################################################
# Class representing script interface to an output
##########################################################################
Oc_Class Oxs_Output {
    private array common closers
    private array common index
    private array common directory
    # Map from the "type" of data provided by an Output to the protocol
    # needed in order to send that data type out.
    private array common protocol {
	"vector field"	vectorField
	DataTable	DataTable
    }

    # We assume that all Oxs_Output objects existing at the same time
    # are provided by the same loaded problem, i.e. the same MIF,
    # so all scalars should be gathered into one DataTable, and one
    # basename is shared in common by all Oxs_Output objects.
    # If Oxs ever allows multiple problems to be loaded at once,
    # re-think this.
    private common scalars {}
    private common basename oxs

    const public variable handle
    const public variable name
    const private variable type
    const public variable units
    ClassConstructor {
	trace variable index wu [subst { uplevel #0 {
	    set opAssocList \[[list array get [$class GlobalName index]]]
	} ;# }]
	array set index [list DataTable DataTable]
	array set directory [list DataTable DataTable]
	foreach d [Oxs_Destination Instances] {
	    if {[string compare DataTable [$d Cget -protocol]] == 0} {
		Oxs_Schedule New _ -output DataTable \
			-destination [$d Cget -name]
	    }
	}
    }
    Constructor {_handle mif} {
	set handle $_handle
	foreach {owner id type units} [Oxs_OutputNames $handle] {break}
	set name "$owner:$id"
	if {[info exists index($name)]} {
	    return -code error "Duplicate output name: $name"
	}
	if {[string match scalar $type]} {
	    if {[lsearch -exact $scalars $name] >= 0} {
		return -code error "Duplicate output name: $name"
	    }
	    lappend scalars $name
	    set scalars [lsort $scalars]
	    set directory($name) $this
	} else {
	    # Register directly with the output list
	    set index($name) $protocol($type)
	    set directory($name) $this
	    foreach d [Oxs_Destination Instances] {
		if {[string compare $index($name) [$d Cget -protocol]] == 0} {
		    Oxs_Schedule New _ -output $name \
			    -destination [$d Cget -name]
		}
	    }
	}

	# Get filename base from the Mif object.
	if {[lsearch -exact [$mif GetMiscKeys] basename] >= 0} {
	    set basename [$mif GetMiscValue basename]
	}

	# When the problem is released, our handle becomes invalid,
	# so we must be destroyed.
	Oc_EventHandler New _ Oxs Release [list $this Delete] \
		-groups [list $this]
    }
    Destructor {
	Oc_EventHandler Generate $this Delete
	Oc_EventHandler DeleteGroup $this
	set idx [lsearch -exact $scalars $name]
	set scalars [lreplace $scalars $idx $idx]
	catch {unset index($name)}
	catch {unset directory($name)}
    }
    proc Lookup {n} {
	return $directory($n)
    }
    private proc SendDataTable {thread} {
	set triples [list [list @Filename:$basename.odt {} 0]]
	foreach on $scalars {
	    set o $directory($on)
	    set n [$o Cget -name]
	    set u [$o Cget -units]
	    set h [$o Cget -handle]
	    lappend triples [list $n $u [Oxs_OutputGet $h]]
	}

	# We're sending DataTable data.  Set up an event to guarantee
	# the table gets closed when the problem ends.
	if {![info exists closers($thread)]} {
	    Oc_EventHandler New closers($thread) Oxs Cleanup \
		    "[list $thread Send DataTable [list [list @Filename: \
		    {} 0]]]; [list unset [$class GlobalName \
		    closers]($thread)]" -groups [list $thread] -oneshot 1
	}
	return [$thread Send DataTable $triples]
    }
    proc Send {n dest} {
	if {[catch {set directory($n)} output]} {
	    return -code error "Unknown output: $n"
	}
	if {[catch {Oxs_Destination Thread $dest} thread]} {
	    return -code error "Unknown destination: $dest"
	}
	
	# DataTable output gets special handling
	if {[string compare DataTable $output]==0} {
	    return [$class SendDataTable $thread]
	}
	return [$output Send $thread]
    }
    method Protocol {} {
	if {[info exists protocol($type)]} {
	    return $protocol($type)
	} 
	return ""
    }
    method Send {thread} {
	# Send my output to $thread.
	# Exact mechanism depends on the protocol
	if {![info exists protocol($type)]} {
	    return -code error "No known protocol for sending output\
		    of type $type"
	}
	switch -exact -- $protocol($type) {
	    vectorField {
		# Piece together a permanent file name
		set idx [lsearch -glob $scalars *Iteration]
		set iter [format %07d [Oxs_OutputGet \
			[$directory([lindex $scalars $idx]) Cget -handle]]]
		set idx [lsearch -glob $scalars *Stage]
		set stage [format %02d [Oxs_OutputGet \
			[$directory([lindex $scalars $idx]) Cget -handle]]]
		switch -exact -- $units {
		    T {set ext .obf ;# vector field in Tesla}
		    A/m {
			switch -glob -- [string tolower $name] {
			    *field* {set ext .ohf}
			    *magnetization* {set ext .omf}
		    	    default {set ext .ovf}
			}
		    }
		    default {set ext .ovf}
		}
		set filename $basename-$name-$stage-$iter$ext
		regsub -all {:+} $filename {-} filename
	        regsub -all "\[ \r\n\t]+" $filename {_} filename

		# Piece together a temporary file name
		Oc_TempFile New t -stem $basename -extension $ext
		set temp [$t AbsoluteName]
		$t Delete
		Oxs_OutputGet $handle $temp
		return [$thread Send datafile [list $temp $filename]]
	    }
	    default {
		return -code error "Don't know how to send output\
			using \"$protocol($type)\# protocol"
	    }
	}
    }
}

##########################################################################
# Class representing a scheduled event
##########################################################################
Oc_Class Oxs_Schedule {

    private array common index

    # Add an interface to change these
    private common events {Step Stage}

    # Is this 
    private array variable active
    private array variable frequency

    const public variable output
    const public variable destination

    Constructor {args} {
	eval $this Configure $args
	# Only one schedule per output, destination pair
	if {![catch {set index($output,$destination)} exists]} {
	    $this Delete
	    array set a $args
	    set index($a(-output),$a(-destination)) $exists
	    return $exists
	}
	set index($output,$destination) $this
	regsub -all {:+} $output : op
	regsub -all "\[ \r\t\n]+" $op _ op
	regsub -all {:+} $destination : d
	regsub -all "\[ \r\t\n]+" $d _ d
	trace variable active wu [subst { uplevel #0 {
	    set "Schedule-$op-$d-active" \
	    \[[list array get [$this GlobalName active]]]
	} ;# }]
	trace variable frequency wu [subst { uplevel #0 {
	    set "Schedule-$op-$d-frequency" \
	    \[[list array get [$this GlobalName frequency]]]
	} ;# }]
	foreach e $events {
	    set active($e) 0
	    set frequency($e) 1
	}

	# Remove the Schedule whenever output or destination goes away.
#	Oc_EventHandler New _ [Oxs_Output Lookup $output] Delete \
#		[list $this Delete] -groups [list $this]
	Oc_EventHandler New _ [Oxs_Destination Lookup $destination] Delete \
		[list $this Delete] -groups [list $this]
    }
    Destructor {
	Oc_EventHandler Generate $this Delete
	Oc_EventHandler DeleteGroup $this
	catch {unset index($output,$destination)}
    }
    proc Set {o d w e v} {
	$index($o,$d) Set$w $e $v
    }
    method Send {e} {
	upvar #0 [string tolower $e] count
	if {$count % $frequency($e)} {
	    return
	}
	if {[catch {Oxs_Output Send $output $destination} msg]} {
	    if {[catch {Oxs_Output Lookup $output}]} {
		$this Delete
	    } else {
		return -code error $msg
	    }
	}
    }
    method SetActive {event value} {
	set active($event) $value
	Oc_EventHandler DeleteGroup $this-$event
	if {$value} {
	    Oc_EventHandler New _ Oxs $event [list $this Send $event] \
		    -groups [list $this-$event $this]
	}
    }
    method SetFrequency {event value} {
	set frequency($event) $value
    }
}

##########################################################################
# Here's the guts of OXSII -- a switchboard between interactive GUI events
# and the set of Tcl commands provided by OXS
##########################################################################
# OXS provides the following classes:
#	FileLogger
#	Oxs_Mif
# and OXS provides the following classes:
#	Oxs_ProbInit $mif $file		: Calls Oxs_ProbRelease
#					: Reads $file into $mif
#					: Creates Ext and Output objects
#					: Resets objects (driver)
#	Oxs_ProbReset			: Resets objects (driver)
#	Oxs_Run				: driver takes steps, may generate
#					: Step, Stage, Done events
#	Oxs_ProbRelease			: Destroys Ext and Output objects
#	Oxs_ListRegisteredExtClasses	: returns names
#	Oxs_ListExtObjects		: returns names
#	Oxs_ListEnergyObjects		: returns names
#	Oxs_ListOutputObjects		: returns output tokens
#	Oxs_OutputGet $outputToken $args	: return or write output value
#	Oxs_OutputNames $outputToken	: returns name/type/units of output
#
# Some of these commands only make sense when the solver is in a
# particular state.  For example, [Oxs_Run] can only succeed if a
# problem is loaded.  So, this script has to keep track of the
# state of the solver.
##########################################################################

# Any changes to status get channeled through [ChangeStatus]
set status UNINITIALIZED
trace variable status w [list ChangeStatus $status]
proc ChangeStatus {old args} {
    global status stage step interface_state problem
    if {[string match $old $status]} {return}
    trace vdelete status w [list ChangeStatus $old]
    trace variable status w [list ChangeStatus $status]
    Oc_EventHandler DeleteGroup ChangeStatus
    switch -exact -- $status {
	"" {
	    # The initial state -- no problem loaded.
	    # Also the state after any problem load fails, or
	    # a problem is released.
	    set stage ""
	    set step ""
	    set interface_state disabled
	}
	Loading... {
	    set stage ""
	    set step ""
	    set interface_state disabled
	    # Let interface get updated with above changes, then
	    # call ProblemLoad
	    after idle [list LoadProblem $problem]
	}
	Initializing... {
	    set stage ""
	    set step ""
	    set interface_state disabled
	    after idle Reset
	}
	Pause {
	    # A problem is loaded, but not running.
	    set interface_state normal
	}
	Run {
	    after idle Loop Run
	}
	Relax {
	    Oc_EventHandler New _ Oxs Stage [list set status Pause] \
		    -oneshot 1 -groups [list ChangeStatus]
	    after idle Loop Relax
	}
	Step {
	    after idle "catch {Oxs_Run}; set status Pause"
	}
	Done {
	    # do nothing ?
	}
	default {error "Status: $status not yet implemented"}
    }
}
# Initialize status to "" -- no problem loaded.
set status ""

# Be sure any loaded problem gets release on application exit
Oc_EventHandler New _ Oc_Main Exit ReleaseProblem
proc ReleaseProblem {} {
    global status
    # We're about to release any loaded problem.  Spread the word.
    Oc_EventHandler Generate Oxs Release
    if {[catch {
	Oxs_ProbRelease
    } msg]} {
	# This is really bad.  Kill the solver.
	Oc_Log Log "Oxs_ProbRelease FAILED:\n\t$msg" panic
	exit
    }
    set status ""
}

proc LoadProblem {f} {
    global status stage step
    # We're about to release any loaded problem.  Spread the word.
    Oc_EventHandler Generate Oxs Release
    Oxs_Mif New t
    if {[catch {
	Oxs_ProbInit $t $f
    } msg] || [catch {
	foreach o [Oxs_ListOutputObjects] {
	    Oxs_Output New _ $o $t
	}
    } msg]} {
	# The problem has been released
	global errorInfo
	after idle [subst {[list set errorInfo $errorInfo]
		[list Oc_Log Log "Error loading $f:\n\t$msg" error]}]
	set status ""
    } else {
	set stage 0
	set step 0
	set status Pause
    }
    $t Delete
}

proc Reset {} {
    global status stage step
    set status Pause
    Oc_EventHandler Generate Oxs Cleanup
    if {[catch {
	Oxs_ProbReset
    } msg]} {
	global errorInfo
	after idle [subst {[list set errorInfo $errorInfo]
		[list Oc_Log Log "Reset error:\n\t$msg" error]}]
	ReleaseProblem
    } else {
	set stage 0
	set step 0
	set status Pause
    }
}

# Keep taking steps as long as the status is unchanged,
# remaining one of Run, Relax.
proc Loop {type} {
    global status
    if {![string match $type $status]} {return}
    if {[catch {Oxs_Run} msg]} {
Oc_Log Log $msg warning
	set status Pause
    }
    after idle [info level 0]
}

# When the Load... menu item chooses a problem, 
# $problem is the file to load.
trace variable problem w {uplevel #0 set status Loading... ;#}

# Interface monitors on Stage and Step counts
Oc_EventHandler New _ Oxs Step [list set step %step]
Oc_EventHandler New _ Oxs Step [list set stage %stage]

# Terminate Loop when solver is Done
Oc_EventHandler New _ Oxs Done [list set status Done]

# All problem releases should request cleanup first
Oc_EventHandler New _ Oxs Release [list Oc_EventHandler Generate Oxs Cleanup]

##########################################################################
# Our server offers GeneralInterface protocol services -- start it
##########################################################################
Net_Protocol New protocol -name "OOMMF GeneralInterface protocol 0.0"
$protocol AddMessage start HasGui {} {
    return [list start [list 0 2]]
}
$protocol AddMessage start GetGui {} \
	[list return [list start [list 0 $gui]]]
$protocol AddMessage start Eval {cmd args} {
    set code [catch {uplevel #0 $args} result]
    if {[llength $cmd]} {
	lappend cmd $code $result
	if {$code == 1} {
	    global errorInfo errorCode
	    lappend cmd $errorInfo $errorCode 
	}
	$connection Tell Result $cmd
    }
    return [list start [list 0 ""]]
}
$protocol AddMessage start Share {local remote trace} {
    global $local

    # The script to be run in the client, when $local is written here
    # %s holds the place for the [set] command to go along with the
    # new value.
    set ctraceFmt "[list trace vdelete $remote w $trace]; %s;\
	    [list trace variable $remote w $trace] ;#"

    # A common prefix for the local write trace, and the script which
    # only starts write traces in the client.  The latter is used
    # when the variable $local does not exist.
    set pfx "[list $connection Tell Result] \[[list format $ctraceFmt]"

    # The script to be run here, when $local is written.  It substitutes
    # the [set] command into $ctraceFmt, then sends the script off to be
    # evaluated in the client.
    set strace "$pfx \[[list list set $remote] \[[list uplevel #0\
	    set $local]]]] ;#"

    # If $local already has a value; initialize the client's copy to
    # also have that value.  If not, just set write traces in the client.
    if {[info exists $local]} {eval $strace} else {eval "$pfx {}]"}

    # Set the trace, so changes here get reflected in the client.
    trace variable $local w $strace

    # Make sure this write trace is destroyed if we ever lose contact
    # with the client
    regsub -all -- % $strace %% strace
    Oc_EventHandler New _ $connection Delete \
	    [list trace vdelete $local w $strace] 
    return [list start [list 0 ""]]
}
$protocol AddMessage start delete {} {
    global protocol
    return [list start [list 0 [after 100 $protocol Delete]]]
}
Oc_EventHandler New _ $protocol Delete [list after 2000 exit] -oneshot 1
Net_Server New server -protocol $protocol -alias Oxsii
$server Start 0

##########################################################################
# Track the threads known to the account server
#	code mostly cribbed from mmLaunch 
#	good candidate for library code?
##########################################################################
# Get Thread info from account server:
proc AccountReady {acct} {
    Oc_MakeNice
    set qid [$acct Send threads]
    Oc_EventHandler New _ $acct Reply$qid [list GetThreadsReply $acct] \
        -groups [list $acct]
}
proc GetThreadsReply { acct } {
    # Set up to receive NewThread messages, but only one handler per account
    Oc_EventHandler DeleteGroup GetThreadsReply-$acct
    Oc_EventHandler New _ $acct Readable [list HandleAccountTell $acct] \
            -groups [list $acct GetThreadsReply-$acct]
    set threads [$acct Get]
    Oc_Log Log "Received thread list: $threads" status
    # Create a Net_Thread instance for each element of the returned
    # thread list
    if {![lindex $threads 0]} {
        foreach triple [lrange $threads 1 end] {
            Net_Thread New _ -hostname [$acct Cget -hostname] \
                    -accountname [$acct Cget -accountname] \
                    -pid [lindex $triple 1]
        }
    }
}
# Detect and handle NewThread message from account server
proc HandleAccountTell { acct } {
    set message [$acct Get]
    switch -exact -- [lindex $message 0] {
        newthread {
            NewThread $acct [lindex $message 1]
        }
        default {
          Oc_Log Log "Bad message from account $acct:\n\t$message" status
        }
    }
}
# There's a new thread with id $id, create corresponding local instance
proc NewThread { acct id } {
    Net_Thread New _ -hostname [$acct Cget -hostname] \
            -accountname [$acct Cget -accountname] -pid $id
}
# Delay "nice" of this process until any children are spawned.
Net_Account New a
if {[$a Ready]} {
    AccountReady $a
} else {   
    Oc_EventHandler New _ $a Ready [list AccountReady $a] -oneshot 1
}

vwait forever

